use actix_files::Files;
use actix_web::{
    self,
    body::BoxBody,
    dev::{
        self, HttpServiceFactory, ResourceDef, Service, ServiceFactory, ServiceRequest,
        ServiceResponse,
    },
    error::{
        ErrorBadRequest, ErrorForbidden, ErrorInternalServerError, ErrorNotFound, InternalError,
    },
    get,
    http::{
        header::{HOST, ORIGIN},
        StatusCode,
    },
    middleware::Logger,
    rt::{self, net::TcpStream},
    web::{self, Data},
    App, Error, HttpRequest, HttpResponse, HttpServer, Responder,
};
use actix_ws::{AggregatedMessage, CloseCode, Closed};
use futures_util::{future::LocalBoxFuture, FutureExt, StreamExt};
use log::{info, warn};
use rustls::{
    crypto::{aws_lc_rs, CryptoProvider},
    pki_types::CertificateDer,
    server::ResolvesServerCert,
    sign::CertifiedKey,
};
use serde::Deserialize;
use serde_json::json;
use std::{
    collections::BTreeMap,
    env,
    fs::{self, File},
    future::{ready, Ready},
    io::{self, BufReader},
    sync::{Arc, RwLock},
};
use tokio::{
    io::{AsyncReadExt, AsyncWriteExt},
    select,
    signal::unix::{signal, SignalKind},
    spawn,
    sync::oneshot,
};
use wildmatch::WildMatch;

#[derive(Deserialize)]
pub struct TlsConfig {
    pub private_key_file: String,
    pub certificate_chain_file: String,
}

#[derive(Deserialize)]
pub struct ServerConfig {
    pub listen_port: u16,
    pub bind_address: String,
    pub tls: Option<TlsConfig>,
    pub static_root: String,
    pub muds: Vec<MudConfig<String>>,
}

impl Default for ServerConfig {
    fn default() -> Self {
        Self {
            listen_port: 443,
            bind_address: "::".to_owned(),
            tls: None,
            static_root: "/frontendstatic".to_owned(),
            muds: vec![],
        }
    }
}

#[derive(Deserialize, Clone, Debug)]
pub struct MudConfig<WildType> {
    pub hostname: String,
    pub upstream_mud: String,
    pub banner_to_mud: String,
    pub startup_script_file: String,
    pub allowed_origins: Vec<WildType>,
}

fn with_mudconfig(
    req: &HttpRequest,
    config_data: Data<DynamicConfigLock>,
) -> Result<MudConfig<WildMatch>, Error> {
    match req.headers().get(&HOST) {
        None => Err(ErrorBadRequest("Missing Host header"))?,
        Some(host) => match config_data
            .read()
            .unwrap()
            .host_map
            .get(host.to_str().unwrap_or("invalid"))
        {
            None => Err(ErrorNotFound(
                "No MUD matching your request is currently enabled",
            ))?,
            Some(v) => Ok(v.clone()),
        },
    }
}

#[get("/ws")]
async fn ws(
    config_data: Data<DynamicConfigLock>,
    req: HttpRequest,
    body: web::Payload,
) -> impl Responder {
    let mud = with_mudconfig(&req, config_data)?;
    match req.headers().get(&ORIGIN) {
        None => Err(ErrorForbidden("Missing origin"))?,
        Some(origin) => {
            if !mud
                .allowed_origins
                .iter()
                .any(|o| o.matches(origin.to_str().unwrap_or("invalid")))
            {
                Err(ErrorForbidden("Disallowed origin"))?;
            }
        }
    }

    let (response, mut session, stream) = actix_ws::handle(&req, body)?;
    let mut stream = stream.aggregate_continuations().max_continuation_size(1024);

    let mut tcp_stream: TcpStream = TcpStream::connect(&mud.upstream_mud).await?;
    let subst_banner = mud
        .banner_to_mud
        .replace(
            "%i",
            &req.peer_addr()
                .map(|a| a.ip().to_string())
                .unwrap_or_else(|| "unknown".to_owned()),
        )
        .replace("%n", "\r\n");
    tcp_stream.write_all(subst_banner.as_bytes()).await?;

    let script = fetch_startup_script(&mud.startup_script_file)
        .map_err(|e| InternalError::new(e, StatusCode::INTERNAL_SERVER_ERROR))?;

    rt::spawn(async move {
        if session
            .text(
                json!({
                    "RunLua": script
                })
                .to_string(),
            )
            .await
            .is_err()
        {
            let _ = session.close(Some(CloseCode::Normal.into())).await;
            return;
        }
        let mut readbuf: [u8; 1024] = [0; 1024];
        loop {
            tokio::select! {
                ws_msg = stream.next() => {
                    match ws_msg {
                        None => break,
                        Some(Err(_e)) => break,
                        Some(Ok(AggregatedMessage::Binary(bin))) => {
                            if tcp_stream.write_all(&bin).await.is_err() {
                                break
                            }
                        }

                        Some(Ok(AggregatedMessage::Ping(msg))) => {
                            if let Err(Closed) = session.pong(&msg).await {
                                break
                            }
                        }
                        Some(Ok(_)) => {},
                    }
                },
                tcp_data_len = tcp_stream.read(&mut readbuf) => {
                    match tcp_data_len {
                        Err(_e) => break,
                        Ok(0) => break,
                        Ok(n) =>
                            if let Err(Closed) = session.binary(readbuf[0..n].to_vec()).await {
                                break;
                            }
                    }

                }
            }
        }
        let _ = session.close(Some(CloseCode::Normal.into())).await;
    });

    Ok::<HttpResponse, Error>(response)
}

// We load this on each connection so it can change.
fn fetch_startup_script(filename: &str) -> anyhow::Result<String> {
    Ok(fs::read_to_string(filename)?)
}

fn load_config_file(filename: &str) -> anyhow::Result<ServerConfig> {
    info!("Reading config file {}", filename);
    Ok(toml::from_str(&fs::read_to_string(filename)?)?)
}

fn validate_config_file(config: &ServerConfig) {
    match fs::metadata(&config.static_root).map(|meta| meta.is_dir()) {
        Err(e) => warn!(
            "Error checking static_root {} is valid: {}",
            &config.static_root, e
        ),
        Ok(false) => warn!("static_root {} is not a directory", &config.static_root),
        Ok(true) => {}
    }
    for mud in &config.muds {
        match fetch_startup_script(&mud.startup_script_file) {
            Err(e) => warn!(
                "Error reading startup script file {} for MUD {}: {}",
                mud.startup_script_file, mud.hostname, e
            ),
            Ok(_) => {}
        }
    }
}

#[derive(Debug)]
struct DynamicConfigData {
    host_map: BTreeMap<String, MudConfig<WildMatch>>,
    certified_key: Option<Arc<CertifiedKey>>,
    static_root: String,
}
type DynamicConfigLock = Arc<RwLock<DynamicConfigData>>;

#[derive(Debug)]
struct DynconfigCertResolver(DynamicConfigLock);

impl ResolvesServerCert for DynconfigCertResolver {
    fn resolve(
        &self,
        _client_hello: rustls::server::ClientHello<'_>,
    ) -> Option<Arc<rustls::sign::CertifiedKey>> {
        self.0.read().unwrap().certified_key.as_ref().cloned()
    }
}

#[derive(Clone)]
struct DynStaticFiles(DynamicConfigLock);
impl HttpServiceFactory for DynStaticFiles {
    fn register(self, config: &mut actix_web::dev::AppService) {
        let rdef = if config.is_root() {
            ResourceDef::root_prefix("/")
        } else {
            ResourceDef::prefix("/")
        };

        config.register_service(rdef, None, self, None)
    }
}

impl ServiceFactory<ServiceRequest> for DynStaticFiles {
    type Response = ServiceResponse;
    type Error = Error;
    type Config = ();
    type Service = DynStaticFiles;
    type InitError = ();
    type Future = Ready<Result<Self::Service, Self::InitError>>;

    fn new_service(&self, _: ()) -> Self::Future {
        ready(Ok(self.clone()))
    }
}

impl Service<ServiceRequest> for DynStaticFiles {
    type Response = ServiceResponse<BoxBody>;
    type Error = Error;
    type Future = LocalBoxFuture<'static, Result<Self::Response, Self::Error>>;

    dev::always_ready!();
    fn call(&self, req: ServiceRequest) -> Self::Future {
        let files = Files::new("/", &self.0.read().unwrap().static_root).index_file("index.html");
        Box::pin(async move {
            files
                .new_service(())
                .await
                .map_err(|_| ErrorInternalServerError("Internal service error"))?
                .call(req)
                .await
        })
    }
}

fn make_dynamic_config_data(config: &ServerConfig) -> anyhow::Result<DynamicConfigData> {
    Ok(DynamicConfigData {
        host_map: config
            .muds
            .iter()
            .map(|c| {
                (
                    c.hostname.clone(),
                    MudConfig::<WildMatch> {
                        hostname: c.hostname.clone(),
                        upstream_mud: c.upstream_mud.clone(),
                        banner_to_mud: c.banner_to_mud.clone(),
                        startup_script_file: c.startup_script_file.clone(),
                        allowed_origins: c
                            .allowed_origins
                            .iter()
                            .map(|ao| WildMatch::new(ao))
                            .collect(),
                    },
                )
            })
            .collect::<BTreeMap<String, MudConfig<WildMatch>>>(),
        certified_key: match &config.tls {
            None => None,
            Some(tls) => {
                let certs: Vec<CertificateDer> = rustls_pemfile::certs(&mut BufReader::new(
                    &mut File::open(&tls.certificate_chain_file)?,
                ))
                .collect::<Result<Vec<CertificateDer>, io::Error>>()?;
                let key = rustls_pemfile::private_key(&mut BufReader::new(&mut File::open(
                    &tls.private_key_file,
                )?))?
                .expect("No private key found in private key file");
                let key = aws_lc_rs::default_provider()
                    .key_provider
                    .load_private_key(key)?;
                CryptoProvider::get_default();
                Some(CertifiedKey::new(certs, key).into())
            }
        },
        static_root: config.static_root.clone(),
    })
}

fn rehash_config(config_file: &str, dynconfig_lock: &DynamicConfigLock) -> anyhow::Result<()> {
    let config: ServerConfig = load_config_file(config_file)?;
    validate_config_file(&config);
    let dynconfig: DynamicConfigData = make_dynamic_config_data(&config)?;

    let dcref: &mut DynamicConfigData = &mut dynconfig_lock.write().unwrap();
    *dcref = dynconfig;

    Ok(())
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    env_logger::builder()
        .filter_level(log::LevelFilter::Info)
        .parse_default_env()
        .init();
    let args: Vec<String> = env::args().collect();
    let config_file = match args.get(1) {
        None => "/etc/worldwideportal_server.toml".to_owned(),
        Some(cfg) => cfg.clone(),
    };
    if config_file == "-h" || config_file == "--help" {
        println!("Usage: worldwideportal-server /path/to/config/file.toml");
        return Ok(());
    }

    let config: ServerConfig = load_config_file(&config_file)?;
    validate_config_file(&config);
    let dynconfig: DynamicConfigLock = RwLock::new(make_dynamic_config_data(&config)?).into();

    // This is a double Arc, but actix doesn't provide an alternative.
    let data = Data::new(dynconfig.clone());

    let server_data = data.clone();
    let static_file_svc = DynStaticFiles(dynconfig.clone());
    let server = HttpServer::new(move || {
        let logger = Logger::default();
        App::new()
            .wrap(logger)
            .app_data(server_data.clone())
            .default_service(static_file_svc.clone())
            .service(ws)
    });

    let server = match &dynconfig.read().unwrap().certified_key {
        None => server.bind((config.bind_address.clone(), config.listen_port))?,
        Some(_tls) => {
            let tls_config = rustls::server::ServerConfig::builder()
                .with_no_client_auth()
                .with_cert_resolver(Arc::new(DynconfigCertResolver(dynconfig.clone())));
            server.bind_rustls_0_23(
                (config.bind_address.clone(), config.listen_port),
                tls_config,
            )?
        }
    };

    let mut sighup = signal(SignalKind::hangup())?;
    let (exit_tx, exit_rx) = oneshot::channel::<()>();
    let exit_rx = exit_rx.shared();
    spawn(async move {
        loop {
            select! {
                _ = sighup.recv() => {
                    let _ = rehash_config(&config_file, &dynconfig);
                }
                _ = exit_rx.clone() => {
                    return;
                }
            }
        }
    });

    server.run().await?;
    let _ = exit_tx.send(());
    Ok(())
}
